;schelog.scm
;schelog
;An embedding of Prolog in Scheme
;(c) Dorai Sitaram, dorai@cs.rice.edu, 1989, Rice University
;revised Feb. 1993

(module schelog:)

(provide 'macro
  (lambda ()
    (load (string-append *library* "mbe.scm"))))

;abbreviation

(define call/cc call-with-current-continuation)

;logic variables and their manipulation

(define *ref* "ref")

(define *unbound* '(global$ _))

(define make-ref
  ;makes a fresh unbound ref;
  ;unbound refs point to themselves
  (lambda opt
    (vector *ref*
      (if (null? opt) *unbound*
	(car opt)))))

(define _ make-ref)

(import #f #t _)

(define ref?
  (lambda (r)
    (and (vector? r)
	 (eq? (vector-ref r 0) *ref*))))

(define deref
  (lambda (r)
    (vector-ref r 1)))

(define set-ref!
  (lambda (r v)
    (vector-set! r 1 v)))

(define unbound-ref?
  (lambda (r)
    (eq? (deref r) *unbound*)))

(define unbind-ref!
  (lambda (r)
    (set-ref! r *unbound*)))

;frozen logic vars

(define *frozen* "frozen")

(define freeze-ref
  (lambda (r)
    (make-ref (vector *frozen* r))))

(define thaw-frozen-ref
  (lambda (r)
    (vector-ref (deref r) 1)))

(define frozen-ref?
  (lambda (r)
    (let ((r2 (deref r)))
      (and (vector? r2)
	   (eq? (vector-ref r2 0) *frozen*)))))

;deref a structure completely (except the frozen ones, i.e.)

(define deref*
  (lambda (s)
    (cond ((ref? s)
	   (if (frozen-ref? s) s
	     (deref* (deref s))))
	  ((pair? s) (cons (deref* (car s)) (deref* (cdr s))))
	  ((vector? s)
	   (list->vector (map deref* (vector->list s))))
	  (else s))))

;letref introduces new logic variables

(define *logic-variables* '())

(define-syntax letref
  (syntax-rules (*logic-variables*)
    ((letref (x ...) . e)
     (let ((x (make-ref)) ...)
       (let ((*logic-variables*
	       (append! (list x ...) *logic-variables*)))
	 . e)))))

(import #f #t letref)

;the unify predicate

(define unify
  (lambda (t1 t2)
    (lambda (fk)
      (letrec
	((cleanup-n-fail 
	   (lambda (s)
	     (for-each unbind-ref! s)
	     (fk 'fail)))
	 (unify1 
	   (lambda (t1 t2 s)
	     (cond ((eqv? t1 t2) s)
		   ((ref? t1)
		    (cond ((unbound-ref? t1)
			   (set-ref! t1 t2)
			   (cons t1 s))
			  ((frozen-ref? t1)
			   (cond ((ref? t2)
				  (cond ((unbound-ref? t2)
					 (set-ref! t2 t1)
					 (cons t2 s))
					((frozen-ref? t2)
					 (cleanup-n-fail s))
					(else
					  (unify1 t1 (deref t2) s))))
				 (else (cleanup-n-fail s))))
			  (else (unify1 (deref t1) t2 s))))
		   ((ref? t2) (unify1 t2 t1 s))
		   ((and (pair? t1) (pair? t2))
		    (unify1 (cdr t1) (cdr t2)
		      (unify1 (car t1) (car t2) s)))
		   ((and (vector? t1) (vector? t2))
		    (unify1 (vector->list t1)
		      (vector->list t2) s))
		   (else
		     (for-each unbind-ref! s)
		     (fk 'fail))))))
	(let ((s (unify1 t1 t2 '())))
	  (lambda (d)
	    (cleanup-n-fail s)))))))

(import #f #t (== unify))

;disjunction

(define-syntax %or
  (syntax-rules ()
    ((%or g ...)
     (lambda (__fk)
       (call/cc
	 (lambda (__sk)
	   (call/cc
	     (lambda (__fk)
	       (__sk ((deref* g) __fk))))
	   ...
	   (__fk 'fail)))))))

;conjunction

(define-syntax %and
  (syntax-rules ()
    ((%and g ...)
     (lambda (__fk)
       (let* ((__fk ((deref* g) __fk))
	      ...)
	 __fk)))))

;cut

(define-syntax %cut
  (syntax-rules (!)
    ((%cut g)
     (lambda (__fk)
       (let ((! (lambda (__fk2) __fk)))
	 ((deref* g) __fk))))))

;Prolog-like sugar

(define-syntax rel
  (syntax-rules (!)
    ((rel (v ...) ((a ...) subgoal ...) ...)
     (lambda __fmls
       (lambda (__fk)
	 (letref (v ...)
	   (let ((! (lambda (fk1) __fk)))
	     (call/cc
	       (lambda (__sk)
		 (call/cc
		   (lambda (__fk)
		     (let* ((__fk ((unify __fmls (list a ...))
				   __fk))
			    (__fk ((deref* subgoal) __fk))
			    ...)
		       (__sk __fk))))
		 ...
		 (__fk 'fail))))))))))

(import-syntax #f #t %or %and %cut rel)

;the fail and true preds

(define %fail
  (lambda (fk) (fk 'fail)))

(define %true
  (lambda (fk) fk))

(import #f #t %fail %true)

;for structures ("functors"), use Scheme's list and vector
;functions and anything that's built using them.

;arithmetic

(define-syntax %is
  (syntax-rules (quote)
    ((%is v e)
     (lambda (__fk)
       ((== v (%is (1) e __fk)) __fk)))

    ((%is (1) (quote x) fk) (quote x))
    ((%is (1) (x ...) fk)
     ((%is (1) x fk) ...))
    ((%is (1) x fk)
     (if (and (ref? x) (unbound-ref? x))
	 (fk 'fail) (deref* x)))))
  
(import-syntax #f #t %is)

;defining arithmetic comparison operators

(define make-binary-arithmetic-relation
  (lambda (f)
    (lambda (x y)
      (%is #t (f x y)))))

(define %eq (make-binary-arithmetic-relation =))
(define %gt (make-binary-arithmetic-relation >))
(define %ge (make-binary-arithmetic-relation >=))
(define %lt (make-binary-arithmetic-relation <))
(define %le (make-binary-arithmetic-relation <=))
(define %ne (make-binary-arithmetic-relation
	      (lambda (m n) (not (= m n)))))

(import #f #t %eq %gt %ge %lt %le %ne)

;type predicates

(define constant?
  (lambda (x)
    (cond ((ref? x)
	   (cond ((unbound-ref? x) #f)
		 ((frozen-ref? x) #t)
		 (else (constant? (deref x)))))
	  ((pair? x) #f)
	  ((vector? x) #f)
	  (else #t))))

(define compound?
  (lambda (x)
    (cond ((ref? x) (cond ((unbound-ref? x) #f)
			  ((frozen-ref? x) #f)
			  (else (compound? (deref x)))))
	  ((pair? x) #t)
	  ((vector? x) #t)
	  (else #f))))

(define %constant
  (lambda (x)
    (lambda (fk)
      (if (constant? x) fk (fk 'fail)))))

(define %compound
  (lambda (x)
    (lambda (fk)
      (if (compound? x) fk (fk 'fail)))))

(import #f #t %constant %compound)

;metalogical type predicates

(define var?
  (lambda (x)
    (cond ((ref? x)
	   (cond ((unbound-ref? x) #t)
		 ((frozen-ref? x) #f)
		 (else (var? (deref x)))))
	  ((pair? x) (or (var? (car x)) (var? (cdr x))))
	  ((vector? x) (var? (vector->list x)))
	  (else #f))))

(define %var
  (lambda (x)
    (lambda (fk) (if (var? x) fk (fk 'fail)))))

(define %nonvar
  (lambda (x)
    (lambda (fk) (if (var? x) (fk 'fail) fk))))

(import #f #t %var %nonvar)

; negation of unify

(define make-negation ;basically inlined cut-fail
  (lambda (p)
    (lambda args
      (lambda (fk)
	(if (call/cc
	      (lambda (k)
		((apply p args) (lambda (d) (k #f)))))
	    (fk 'fail)
	    fk)))))

(define %notunify
  (make-negation unify))

(import #f #t %notunify)

(define ident?
  (lambda (x y)
    (cond ((ref? x)
	   (cond ((unbound-ref? x)
		  (cond ((ref? y)
			 (cond ((unbound-ref? y) (eq? x y))
			       ((frozen-ref? y) #f)
			       (else (ident? x (deref y)))))
			(else #f)))
		 ((frozen-ref? x)
		  (cond ((ref? y)
			 (cond ((unbound-ref? y) #f)
			       ((frozen-ref? y) (eq? x y))
			       (else (ident? x (deref y)))))
			(else #f)))
		 (else (ident? (deref x) y))))
	  ((pair? x)
	   (cond ((ref? y)
		  (cond ((unbound-ref? y) #f)
			((frozen-ref? y) #f)
			(else (ident? x (deref y)))))
		 ((pair? y)
		  (and (ident? (car x) (car y))
		       (ident? (cdr x) (cdr y))))
		 (else #f)))
	  ((vector? x)
	   (cond ((ref? y)
		  (cond ((unbound-ref? y) #f)
			((frozen-ref? y) #f)
			(else (ident? x (deref y)))))
		 ((vector? y)
		  (ident? (vector->list x)
		    (vector->list y)))
		 (else #f)))
	  (else
	    (cond ((ref? y)
		   (cond ((unbound-ref? y) #f)
			 ((frozen-ref? y) #f)
			 (else (ident? x (deref y)))))
		  ((pair? y) #f)
		  ((vector? y) #f)
		  (else (eqv? x y)))))))

(define %ident
  (lambda (x y)
    (lambda (fk) (if (ident? x y) fk (fk 'fail)))))

(define %notident
  (lambda (x y)
    (lambda (fk) (if (ident?) (fk 'fail) fk))))

(import #f #t %ident %notident)

;variables as objects

(define freeze
  (lambda (s)
    (let ((dict '()))
      (let loop ((s s))
	(cond ((ref? s)
	       (cond ((or (unbound-ref? s) (frozen-ref? s))
		      (let ((x (assq s dict)))
			(if x (cdr x)
			    (let ((y (freeze-ref s)))
			      (set! dict (cons (cons s y) dict))
			      y))))
		     ;((frozen-ref? s) s) ;?
		     (else (loop (deref s)))))
	      ((pair? s) (cons (loop (car s)) (loop (cdr s))))
	      ((vector? s)
	       (list->vector (map loop (vector->list s))))
	      (else s))))))

(define melt
  (lambda (f)
    (cond ((ref? f)
	   (cond ((unbound-ref? f) f)
		 ((frozen-ref? f) (thaw-frozen-ref f))
		 (else (melt (deref f)))))
	  ((pair? f)
	   (cons (melt (car f)) (melt (cdr f))))
	  ((vector? f)
	   (list->vector (map melt (vector->list f))))
	  (else f))))

(define melt-new
  (lambda (f)
    (let ((dict '()))
      (let loop ((f f))
	(cond ((ref? f)
	       (cond ((unbound-ref? f) f)
		     ((frozen-ref? f)
		      (let ((x (assq f dict)))
			(if x (cdr x)
			    (let ((y (make-ref)))
			      (set! dict (cons (cons f y) dict))
			      y))))
		     (else (loop (deref f)))))
	      ((pair? f) (cons (loop (car f)) (loop (cdr f))))
	      ((vector? f)
	       (list->vector (map loop (vector->list f))))
	      (else f))))))

(define copy
  (lambda (s)
    (melt-new (freeze s))))

(define %freeze
  (lambda (s f)
    (lambda (fk)
      ((unify (freeze s) f) fk))))

(define %melt
  (lambda (f s)
    (lambda (fk)
      ((unify (melt f) s) fk))))

(define %melt-new
  (lambda (f s)
    (lambda (fk)
      ((unify (melt-new f) s) fk))))

(define %copy
  (lambda (s c)
    (lambda (fk)
      ((unify (copy s) c) fk))))

(import #f #t %freeze %melt %melt-new %copy)

;negation as failure

(define %not
  (lambda (g)
    (lambda (fk)
      (if (call/cc
	    (lambda (k)
	      ((deref* g) (lambda (d) (k #f)))))
	  (fk 'fail) fk))))

(import #f #t %not)

;set predicates

(define set-cons
  (lambda (e s)
    (if (member e s) s (cons e s))))

(define every2
  (lambda (p s1 s2)
    (let loop ((s1 s1) (s2 s2))
      (cond ((and (null? s1) (null? s2)) #t)
	    ((or (null? s1) (null? s2)) #f)
	    ((p (car s1) (car s2))
	     (loop (cdr s1) (cdr s2)))
	    (else #f)))))

(define remove-if!
  (lambda (p s)
    (let loop ((s s))
      (cond ((null? s) '())
	    ((p (car s)) (loop (cdr s)))
	    (else (let ((r (loop (cdr s))))
		    (set-cdr! s r)
		    s))))))

(define find-unbounds
  (lambda (e1 . e2)
    (let ((s '()))
      (let loop ((e e2))
	(cond ((null? e) 'skip)
	      ((pair? e)
	       (loop (car e))
	       (loop (cdr e)))
	      ((ref? e)
	       (cond ((unbound-ref? e)
		      (if (memq e s) 'skip
			(set! s (cons e s))))
		     ((frozen-ref? e) 'skip)
		     (else (loop (deref e)))))
	      ((vector? e)
	       (let ((n (vector-length e)))
		 (do ((i 0 (+ i 1)))
		     ((>= i n))
		     (loop (vector-ref e i)))))
	      (else 'skip)))
      (let ((r '()))
	(let loop ((e e1))
	  (cond ((null? e) 'skip)
		((pair? e)
		 (loop (car e))
		 (loop (cdr e)))
		((ref? e)
		 (cond ((unbound-ref? e)
			(if (or (memq e s) (memq e r)) 'skip
			  (set! r (cons e r))))
		       ((frozen-ref? e) 'skip)
		       (else (loop (deref e)))))
		((vector? e)
		 (let ((n (vector-length e)))
		   (do ((i 0 (+ i 1)))
		       ((>= i n))
		       (loop (vector-ref e i)))))
		(else 'skip)))
	r))))

(define restore-unbounds
  (lambda (s)
    (for-each unbind-ref! s)))

(define make-bag-of
  (lambda (kons)
    (lambda (lv fvv goal bag)
      (lambda (fk)
	(let* ((uu (find-unbounds fvv lv bag))
	       (ans
		 (call/cc
		   (lambda (ans-k)
		     (let* ((bag '())
			    (fk-final
			      (lambda (d)
				(let ((s (reverse! bag)))
				  ;merely to gc
				  (set! bag '())
				  (ans-k s))))
			    (fk (goal fk-final)))
		       (set! bag (kons (deref* lv) bag))
		       (fk 'retry))))))
	  (restore-unbounds uu)
	  ((unify ans bag) fk))))))
				
(define bag-of (make-bag-of cons))
(define set-of (make-bag-of set-cons))

(define-syntax %bag-of
  (syntax-rules ()
    ((%bag-of x goal bag)
     (bag-of x *logic-variables* goal bag))))

(define-syntax %set-of
  (syntax-rules ()
    ((%set-of x goal set)
     (set-of x *logic-variables* goal set))))

'(define-syntax %exists
  (syntax-rules ()
    ((%exists y goal)
     (%and (letref (__tmp)
	     (%bag-of y goal __tmp)) %true))))

;but this is probably cheaper

(define-syntax %exists
  (syntax-rules ()
    ((%exists y goal)
     (let ((y (copy y))) goal))))

;%bag-of-1, %set-of-1 hold if there's at least one solution

(define-syntax %bag-of-1
  (syntax-rules ()
    ((%bag-of-1 x g ii)
     (%and (%bag-of x g ii)
       (unify ii (cons (_) (_)))))))

(define-syntax %set-of-1
  (syntax-rules ()
    ((%set-of-1 x g ii)
     (%and (%set-of x g ii)
       (unify ii (cons (_) (_)))))))

(import-syntax #f #t %bag-of %set-of %exists
  %bag-of-1 %set-of-1)

;user interface

;(which (v ...) query) returns #f if query fails and instantiations
;of v ... if query succeeds.  In the latter case, type (more) to
;retry query for more instantiations.

(define *more-k* 'forward)
(define *more-fk* 'forward)

(defmacro which (vv g)
  `(letref ,vv
     (call/cc
       (lambda (__qk)
	 (set! *more-k* __qk)
	 (set! *more-fk*
	   ((deref* ,g) 
	    (lambda (d)
	      (set! *more-fk* #f)
	      (*more-k* #f))))
	 (*more-k*
	   (map (lambda (nam val) (list nam (deref* val)))
		(list ,@(map (lambda (x) `',x) vv))
		(list ,@vv)))))))
		  
(define more
  (lambda ()
    (call/cc
      (lambda (k)
	(set! *more-k* k)
	(if *more-fk* (*more-fk* 'more)
	  #f)))))

(import-syntax #f #t which)
(import #f #t more)

;end of embedding code.  The following are
;some utilities, written in Schelog

(define %member
  (lambda (x y)
    (letref (xs z zs)
      (%or
	(== y (cons x xs))
	(%and (== y (cons z zs))
	  (%member x zs))))))

(define %if-then-else
  (lambda (p q r)
    (%cut
      (%or
	(%and p ! q)
	r))))

;the above could also have been written in a more
;Prolog-like fashion, viz.

'(define %member
  (rel (x xs y ys)
    ((x (cons x xs)))
    ((x (cons y ys)) (%member x ys))))

'(define %if-then-else
  (rel (p q r)
    ((p q r) p ! q)
    ((p q r) r)))

(import #f #t %member %if-then-else)